@@ -41,7 +41,7 @@
 
 #define MODE_TABLE_OFFSET_845G 617
 
-#define VERSION "0.5"
+#define VERSION "0.5.2"
 
 #define ATI_SIGNATURE1 "ATI MOBILITY RADEON"
 #define ATI_SIGNATURE2 "ATI Technologies Inc"
@@ -55,10 +55,10 @@ typedef unsigned char boolean;
 typedef unsigned int cardinal;
 
 typedef enum {
-    CT_UNKWN, CT_845G, CT_855GM, CT_865G, CT_915GM, CT_945G, CT_915G
+    CT_UNKWN, CT_845G, CT_855GM, CT_865G, CT_915G, CT_915GM, CT_945G, CT_945GM 
 } chipset_type;
 
-char * chipset_type_names[] = {"UNKNOWN", "845G", "855GM", "865G", "915GM", "945G", "915G"};
+char * chipset_type_names[] = {"UNKNOWN", "845G", "855GM", "865G", "915G", "915GM", "945G", "945GM"};
 
 typedef enum {
     BT_UNKWN, BT_1, BT_2, BT_3
@@ -78,10 +78,10 @@ typedef struct {
 typedef struct {
     byte unknow1[2];
     byte x1;
-    byte unknow2;
+    byte x_total;
     byte x2;
     byte y1;
-    byte unknow3;
+    byte y_total;
     byte y2;
 } __attribute__((packed)) vbios_resolution_type1;
 
@@ -189,6 +189,10 @@ chipset_type get_chipset(cardinal id) {
         type = CT_865G;
         break;
 
+    case 0x25808086:
+	type = CT_915G;
+	break;
+
     case 0x25908086:
         type = CT_915GM;
         break;
@@ -197,9 +201,9 @@ chipset_type get_chipset(cardinal id) {
         type = CT_945G;
         break;
 
-    case 0x25808086:
-	type = CT_915G;
-	break;
+    case 0x27a08086:
+        type = CT_945GM;
+        break;
 
     default:
         type = CT_UNKWN;
@@ -480,9 +484,10 @@ void unlock_vbios(vbios_map * map) {
             break;
         case CT_845G:
         case CT_865G:
+	case CT_915G:
         case CT_915GM:
         case CT_945G:
-	case CT_915G:
+        case CT_945GM:
             outl(0x80000090, 0xcf8);
             map->b1 = inb(0xcfd);
             map->b2 = inb(0xcfe);
@@ -526,9 +531,10 @@ void relock_vbios(vbios_map * map) {
             break;
         case CT_845G:
         case CT_865G:
+	case CT_915G:
         case CT_915GM:
         case CT_945G:
-	case CT_915G:
+        case CT_945GM:
             outl(0x80000090, 0xcf8);
             outb(map->b1, 0xcfd);
             outb(map->b2, 0xcfe);
@@ -552,7 +558,7 @@ void relock_vbios(vbios_map * map) {
 }
 
 
-void list_modes(vbios_map *map) {
+void list_modes(vbios_map *map, cardinal raw) {
     cardinal i, x, y;
 
     for (i=0; i < map->mode_table_size; i++) {
@@ -567,6 +573,12 @@ void list_modes(vbios_map *map) {
                 if (x != 0 && y != 0) {
                     printf("Mode %02x : %dx%d, %d bits/pixel\n", map->mode_table[i].mode, x, y, map->mode_table[i].bits_per_pixel);
                 }
+
+		if (raw)
+		{
+                    printf("Mode %02x (raw) :\n\t%02x %02x\n\t%02x\n\t%02x\n\t%02x\n\t%02x\n\t%02x\n\t%02x\n", map->mode_table[i].mode, res->unknow1[0],res->unknow1[1], res->x1,res->x_total,res->x2,res->y1,res->y_total,res->y2);
+		}
+
             }
             break;
         case BT_2:
@@ -599,7 +611,7 @@ void list_modes(vbios_map *map) {
     }
 }
 
-void set_mode(vbios_map * map, cardinal mode, cardinal x, cardinal y, cardinal bp) {
+void set_mode(vbios_map * map, cardinal mode, cardinal x, cardinal y, cardinal bp, cardinal htotal, cardinal vtotal) {
     cardinal i, j;
 
     for (i=0; i < map->mode_table_size; i++) {
@@ -613,11 +625,16 @@ void set_mode(vbios_map * map, cardinal mode, cardinal x, cardinal y, cardinal b
                         map->mode_table[i].bits_per_pixel = bp;
                     }
                     
-                    res->x2 = (res->x2 & 0x0f) | ((x >> 4) & 0xf0);
+                    res->x2 = (htotal?(((htotal-x) >> 8) & 0x0f) : (res->x2 & 0x0f)) | ((x >> 4) & 0xf0);
                     res->x1 = (x & 0xff);
                     
-                    res->y2 = (res->y2 & 0x0f) | ((y >> 4) & 0xf0);
+                    res->y2 = (vtotal?(((vtotal-y) >> 8) & 0x0f) : (res->y2 & 0x0f)) | ((y >> 4) & 0xf0);
                     res->y1 = (y & 0xff);
+		    if (htotal)
+			res->x_total = ((htotal-x) & 0xff);
+
+		    if (vtotal)
+			res->y_total = ((vtotal-y) & 0xff);
                 }
                 break;
             case BT_2:
@@ -630,6 +647,11 @@ void set_mode(vbios_map * map, cardinal mode, cardinal x, cardinal y, cardinal b
                         if (modeline->x1 == res->modelines[0].x1) {
                             modeline->x1 = modeline->x2 = x-1;
                             modeline->y1 = modeline->y2 = y-1;
+			    if (htotal)
+                        	modeline->htotal = htotal;
+
+			    if (vtotal)
+                        	modeline->vtotal = vtotal;
                             
                             /* WARNING: I don't know at all what are the formulas
                                to calculate these values !
@@ -639,11 +661,9 @@ void set_mode(vbios_map * map, cardinal mode, cardinal x, cardinal y, cardinal b
                             
                             modeline->hsyncstart = x+10;
                             modeline->hsyncend   = x+20;
-                            modeline->htotal     = x+30;
                             
                             modeline->vsyncstart = y+10;
                             modeline->vsyncend   = y+20;
-                            modeline->vtotal     = y+30;
                         }
                     }
                 }
@@ -659,6 +679,12 @@ void set_mode(vbios_map * map, cardinal mode, cardinal x, cardinal y, cardinal b
                             modeline->x1 = modeline->x2 = x-1;
                             modeline->y1 = modeline->y2 = y-1;
                             
+			    if (htotal)
+                        	modeline->htotal = htotal;
+
+			    if (vtotal)
+                        	modeline->vtotal = vtotal;
+
                             /* WARNING: I don't know at all what are the formulas
                                to calculate these values !
                                The following lines are only an example.
@@ -668,11 +694,9 @@ void set_mode(vbios_map * map, cardinal mode, cardinal x, cardinal y, cardinal b
                             
                             modeline->hsyncstart = x+10;
                             modeline->hsyncend   = x+20;
-                            modeline->htotal     = x+30;
                             
                             modeline->vsyncstart = y+10;
                             modeline->vsyncend   = y+20;
-                            modeline->vtotal     = y+30;
                             
                             modeline->timing_h   = 0x2ff;
                             modeline->timing_v   = 0x4ff;
@@ -696,10 +720,10 @@ void display_map_info(vbios_map * map) {
 }
 
 
-int parse_args(int argc, char *argv[], char ** filename, chipset_type *forced_chipset, cardinal *list, cardinal *mode, cardinal *x, cardinal *y, cardinal *bp) {
+int parse_args(int argc, char *argv[], char ** filename, chipset_type *forced_chipset, cardinal *list, cardinal *mode, cardinal *x, cardinal *y, cardinal *bp, cardinal *raw, cardinal *htotal, cardinal *vtotal) {
     cardinal index = 1;
 
-    *list = *mode = *x = *y = 0;
+    *list = *mode = *x = *y = *raw = *htotal = *vtotal = 0;
 
     *forced_chipset = CT_UNKWN;
     
@@ -737,19 +761,25 @@ int parse_args(int argc, char *argv[], char ** filename, chipset_type *forced_ch
         else if (!strcmp(argv[index], "865")) {
             *forced_chipset = CT_865G;
         }
-        else if (!strcmp(argv[index], "915")) {
+        else if (!strcmp(argv[index], "915G")) {
+            *forced_chipset = CT_915G;
+        }
+        else if (!strcmp(argv[index], "915GM")) {
             *forced_chipset = CT_915GM;
         }
-        else if (!strcmp(argv[index], "945")) {
+        else if (!strcmp(argv[index], "945G")) {
             *forced_chipset = CT_945G;
         }
+        else if (!strcmp(argv[index], "945GM")) {
+            *forced_chipset = CT_945GM;
+        }
         else {
             *forced_chipset = CT_UNKWN;
         }
         
         index++;
         
-        if(argc<=index) {
+        if (argc<=index) {
             return 0;
         }
     }
@@ -763,7 +793,16 @@ int parse_args(int argc, char *argv[], char ** filename, chipset_type *forced_ch
         }
     }
     
-    if (argc-index != 3 && argc-index != 4) {
+    if ((argc > index) && !strcmp(argv[index], "-r")) {
+        *raw = 1;
+        index++;
+
+        if(argc<=index) {
+            return 0;
+        }
+    }
+    
+    if (argc-index < 3 || argc-index > 6) {
         return -1;
     }
 
@@ -772,34 +811,50 @@ int parse_args(int argc, char *argv[], char ** filename, chipset_type *forced_ch
     *y = (cardinal)atoi(argv[index+2]);
 
 
-    if (argc-index == 4) {
+    if (argc-index > 3) {
         *bp = (cardinal)atoi(argv[index+3]);
     }
     else {
         *bp = 0;
     }
     
+    if (argc-index > 4) {
+        *htotal = (cardinal)atoi(argv[index+4]);
+    }
+    else {
+        *htotal = 0;
+    }
+
+    if (argc-index > 5) {
+        *vtotal = (cardinal)atoi(argv[index+5]);
+    }
+    else {
+        *vtotal = 0;
+    }
+    
     return 0;
 }
 
 void usage(char *name) {
-    printf("Usage: %s [-f file] [-c chipset] [-l] [mode X Y] [bits/pixel]\n", name);
+    printf("Usage: %s [-f file] [-c chipset] [-l] [mode X Y] [bits/pixel] [htotal] [vtotal]\n", name);
     printf("  Set the resolution to XxY for a video mode\n");
+    printf("  Bits per pixel are optional.  htotal/vtotal settings are additionally optional.\n");
     printf("  Options:\n");
     printf("    -f use an alternate file (THIS IS USED FOR DEBUG PURPOSES)\n");
     printf("    -c force chipset type (THIS IS USED FOR DEBUG PURPOSES)\n");
     printf("    -l display the modes found in the video BIOS\n");
+    printf("    -r display the modes found in the video BIOS in raw mode (THIS IS USED FOR DEBUG PURPOSES)\n");
 }
 
 int main (int argc, char *argv[]) {
     vbios_map * map;
-    cardinal list, mode, x, y, bp;
+    cardinal list, mode, x, y, bp, raw, htotal, vtotal;
     char * filename;
     chipset_type forced_chipset;
     
     printf("Intel 800/900 Series VBIOS Hack : version %s\n\n", VERSION);
 
-    if (parse_args(argc, argv, &filename, &forced_chipset, &list, &mode, &x, &y, &bp) == -1) {
+    if (parse_args(argc, argv, &filename, &forced_chipset, &list, &mode, &x, &y, &bp, &raw, &htotal, &vtotal) == -1) {
         usage(argv[0]);
         return 2;
     }
@@ -812,18 +867,18 @@ int main (int argc, char *argv[]) {
     printf("\n");
 
     if (list) {
-        list_modes(map);
+        list_modes(map, raw);
     }
 
     if (mode!=0 && x!=0 && y!=0) {
         unlock_vbios(map);
-        set_mode(map, mode, x, y, bp);
+        set_mode(map, mode, x, y, bp, htotal, vtotal);
         relock_vbios(map);
         
         printf("Patch mode %02x to resolution %dx%d complete\n", mode, x, y);
         
         if (list) {
-            list_modes(map);
+            list_modes(map, raw);
         }
     }
 
